//**********************************************************************
//  state_machine_attempt5.ino
//  Version   YY/MM/DD     Comments
//  5.00      22/07/11     Controls the Twitch Operated (internet) Accessible Delivery Robot
//**********************************************************************

//**********************************************************************
//                        INCLUDING LIBRARIES
//**********************************************************************
#include <ESP8266WiFi.h>
#include <IRCClient.h>
#include <NewPing.h>


//**********************************************************************
//                            TWITCH SETUP
//**********************************************************************

// NETWORKING
//
#define IRC_SERVER   "irc.chat.twitch.tv"
#define IRC_PORT     6667

char ssid[] = "iPhone";       // your network SSID (name)
char password[] = "scotchegg";  // your network key

const String twitchChannelName = "ctrlardu"; //The name of the channel that you want the bot to join
#define TWITCH_BOT_NAME "exampleName" //The name that you want the bot to have
#define TWITCH_OAUTH_TOKEN "oauth:bp7697kw1tf41d4ich0qha7gos6igt" //OAuth Key for your twitch bot
// https://twitchapps.com/tmi/

String ircChannel = "";

WiFiClient wiFiClient;
IRCClient client(IRC_SERVER, IRC_PORT, wiFiClient);

//**********************************************************************
//           SETTING PINS AND SETTING GLOBAL VARIABLES
//**********************************************************************

// Motor Pins
//
int ENA = D7;
int IN1 = D1;
int IN2 = D2;

int ENB = D5;
int IN3 = D3;
int IN4 = D4;

// Global Variables
//
#define SOUND_SPEED 0.034
#define CM_TO_INCH 0.393701
float distanceCm;
float distanceInch;

String input = "";
char numberinput = ' ';
int timestart = 0;
float DEGREE = 0;
long duration;

int TURN_DELAY = 375;
int MOVE_DELAY = 500;
int TALLY_DELAY = 3000;
int tallytimer = 0;

int totaltime = 0;

static enum STATE{ STANDBY, FORWARD, BACKWARD, LEFT, RIGHT, RANDOM } state = STANDBY;
String statename[] = { "Standby", "Forward", "Backward", "Left", "Right", "Random" };
int randumb[] = {0, 1, 2, 3};
int twice = 0;

// Tally Variables
//
uint8_t countTally[] = {0, 0, 0, 0, 0, 0};
uint8_t a;
uint8_t maxTally = countTally[0];

int ind = 0;
uint8_t n = sizeof(countTally);
bool tieVar = false;

// Dist Sensor Pins and Variables
//
#define trigPin 15
#define echoPin 12
#define maxdist 400
NewPing sonar(trigPin, echoPin, maxdist);
int DISTANCE = 0;


//**********************************************************************
//                     SETUP AND FUNCTION DEFINITION
//**********************************************************************

void setup() {

  // Serial Communication for Debugging
  //
  Serial.begin(115200);
  Serial.println();

  // Set WiFi to station mode and disconnect from an AP if it was Previously
  // connected
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  delay(100);

  // Attempt to connect to Wifi network:
  //
  Serial.print("Connecting Wifi: ");
  Serial.println(ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print(".");
    delay(500);
  }
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  IPAddress ip = WiFi.localIP();
  Serial.println(ip);

  ircChannel = "#" + twitchChannelName;

  client.setCallback(callback);

  // Distance Sensor Pin Setup
  //
  pinMode(trigPin, OUTPUT); // Sets the trigPin as an Output
  pinMode(echoPin, INPUT); // Sets the echoPin as an Input

  // Setup Motor Driver
  //
  pinMode(ENA, OUTPUT); //right motor (front is at swivel wheel)
  pinMode(IN1, OUTPUT);
  pinMode(IN2, OUTPUT);
  digitalWrite(IN1, LOW);
  digitalWrite(IN2, LOW);

  pinMode(ENB, OUTPUT); //left motor
  pinMode(IN3, OUTPUT);
  pinMode(IN4, OUTPUT);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4, LOW);

  // Start Timer for Tally Period
  //
  tallytimer = millis();
}

// MAIN LOOP
//
void loop() {
  // Try to connect to chat. If it loses connection try again
  if (!client.connected())
  {
    Serial.println("Attempting to connect to " + ircChannel );
    if (client.connect(TWITCH_BOT_NAME, "", TWITCH_OAUTH_TOKEN)) {
      client.sendRaw("JOIN " + ircChannel);
      Serial.println("connected and ready to rock");
      sendTwitchMessage("Ready to go Boss!");
    }
    else
    {
      Serial.println("failed... try again in 5 seconds");
      // Wait 5 seconds before retrying
      delay(5000);
    }
    return;
  }
  // Summon the State Machine!!!
  sm_01();
  client.loop();
}

// FUNCTION TO SEND A TWITCH MESSAGE
//
void sendTwitchMessage(String message) {
  client.sendMessage(ircChannel, message);
}

// CALLBACK WHENEVER A TWITCH MESSAGE IS SENT IN CHAT
//
void callback(IRCMessage ircMessage) {
  //Serial.println("In CallBack");

  if (ircMessage.command == "PRIVMSG" && ircMessage.text[0] != '\001') {
    //Serial.println("Passed private message.");
    ircMessage.nick.toUpperCase();

    String message = ("<" + ircMessage.nick + "> " + ircMessage.text);
    input = ircMessage.text;

    if (input == "forward")
    {
      countTally[1]++;
    }

    else if (input == "backward")
    {
      countTally[2]++;
    }

    else if (input == "turn left")
    {
      countTally[3]++;
    }

    else if (input == "turn right")
    {
      countTally[4]++;
    }

    else if (input == "random")
    {
      countTally[5]++;
    }

  }
}

// READING DISTANCE FUNCTION
//
int readDistance() {
  unsigned int uS = sonar.ping();
  if (uS == 0)
  {
    pinMode(echoPin, OUTPUT);
    delay(150);
    digitalWrite(echoPin, LOW);
    delay(150);
    pinMode(echoPin, INPUT);
    delay(150);
    return 1000;
  }
  else
  {
    distanceCm = sonar.convert_cm(uS);
    return distanceCm;
  }
}

// STOPPING FUNCTION
//
void stopper()
{
  analogWrite(ENA, 0);
  digitalWrite(IN1, LOW);
  digitalWrite(IN2, LOW);

  analogWrite(ENB, 0);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4, LOW);
  delay(100);
  return;
}

//**********************************************************************
//                      STATE MACHINE FUNCTION
//**********************************************************************

void sm_01()
{
  switch (state)
  {
    case STANDBY:
    // Serial.println("Case Standby");
      stopper();

      if(millis() - tallytimer > TALLY_DELAY)
      {
        for (auto& a : countTally) //range-based for loop
        {
          if (a > maxTally)
          {
            maxTally = a;
          }
        }
  
        while (ind < n)
        {
          if (countTally[ind] == maxTally) 
          {
            break;
          }
          ind++;
        }
  
        for (int count = 0; count < ind; count++)
        {
          if(countTally[count] == maxTally)
          {
            tieVar = true;
            /* if(tieVar)
            {
             Serial.println("Tie1!");
            }
            else
            {
            Serial.println("No Tie1!");
            } */ 
            break;
          }
        }
        for (int count = 5; count > ind; count--)
        {
          if(countTally[count] == maxTally)
          {
            tieVar = true;
            /* if(tieVar)
            {
             Serial.println("Tie2!");
            }
            else
            {
            Serial.println("No Tie2!");
            } */
            break;
          }
        }
  
        if(tieVar != true) 
        {
          state = (STATE)ind;
          sendTwitchMessage("Votes Tallied!");
          sendTwitchMessage(statename[ind]);
          tallytimer = 0;
          timestart = millis();
          memset(countTally, 0, sizeof(countTally));
          maxTally = countTally[0];
          a = 0;
          ind = 0;
        }
        else
        {
          tallytimer = millis();
          a = 0;
          ind = 0;
          maxTally = countTally[0];
          tieVar = false;
        }
      }

      break;
      

    case FORWARD:   // wait for the delay period
      // Serial.print(timestart);
      Serial.println("Forward");
      analogWrite(ENA, 255);
      analogWrite(ENB, 179);
      digitalWrite(IN1, LOW);
      digitalWrite(IN2, HIGH);
      digitalWrite(IN3, HIGH);
      digitalWrite(IN4, LOW);

      DISTANCE = readDistance();
      totaltime =  millis() - timestart;
      if (DISTANCE < 15 || totaltime >= MOVE_DELAY)
      {
        stopper();
        state = STANDBY;
        timestart = 0;
        tallytimer = millis();
        Serial.println(DISTANCE);
        break;
      }
      delay(10);
      break;

    case BACKWARD:   // wait for the delay period
      // Serial.print(timestart);
      analogWrite(ENA, 255);
      analogWrite(ENB, 179);
      digitalWrite(IN1, HIGH);
      digitalWrite(IN2, LOW);
      digitalWrite(IN3, LOW);
      digitalWrite(IN4, HIGH);


      delay(MOVE_DELAY);
      state = STANDBY;
      delay(20);
      timestart = 0;
      tallytimer = millis();
      break;

    case LEFT:   // wait for the delay period
      //Serial.print(timestart);
      analogWrite(ENA, 255);
      analogWrite(ENB, 179);
      digitalWrite(IN1, LOW);
      digitalWrite(IN2, HIGH);
      digitalWrite(IN3, LOW);
      digitalWrite(IN4, HIGH);

      DISTANCE = readDistance();
      totaltime =  millis() - timestart;
      if (totaltime >= TURN_DELAY)
      {
        Serial.print("stop!");
        stopper();
        state = STANDBY;
        timestart = 0;
        tallytimer = millis();
        Serial.println(DISTANCE);
        break;
      }
      delay(10);
      break;

    case RIGHT:   // wait for the delay period
      // Serial.print(timestart);
      analogWrite(ENA, 255);
      analogWrite(ENB, 179);
      digitalWrite(IN1, HIGH);
      digitalWrite(IN2, LOW);
      digitalWrite(IN3, HIGH);
      digitalWrite(IN4, LOW);

      DISTANCE = readDistance();
      totaltime =  millis() - timestart;
      if (totaltime >= TURN_DELAY)
      {
        Serial.print("stop!");
        stopper();
        state = STANDBY;
        timestart = 0;
        tallytimer = millis();
        Serial.println(DISTANCE);
        break;
      }
      delay(10);
      break;

    case RANDOM: //KEEP THIS AT THE END
      int numba = rand() % 4;
      switch (randumb[numba])
      {

        case 0:
          state = FORWARD;
          break;

        case 1:
          state = BACKWARD;
          break;

        case 2:
          state = LEFT;
          break;

        case 3:
          state = RIGHT;
          break;
      }
      break;
  }
}
